#if 0
Delete -i mkfilter
If "{1}"=="-apptool";C -o mkfilter mkfilter.o -i :: -t 'APPL' -c "????" -d APPTOOL
Else;C -o mkfilter mkfilter.o -i ::;End
Exit 0
#endif
/*
 mkfilter.c
  polyphase filter generator

  Created: 2018.10.30
  Revised: 2022.11.05
   Author: Adam Pickett

2018.11.21

December 14, 2021
 merged dolph.c into mkfilter.
October 16, 2022
 added frequency response measuring code.
*/

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#if TARGET_OS_MAC /* MacRoman */
 #include <fp.h>
 //#define CHAR_PI		'\271'
 //#define CHAR_DIVIDE		'\326'
#else
 #include <math.h>
 #include <io.h>
 #include <fcntl.h>
 typedef double	double_t;
#endif

#ifndef M_PI
/* the standard library is inaccurate to the value of pi */
/* partly because it reads long decimal numbers wrong */
/* adding the integer portion separately increases accuracy */
/* but w/o recoding sin etc., it helps virtually nothing */
 #define M_PI			3.1415926535897932384626433832795
#endif
#define MIN(a,b)		((a)<(b)?(a):(b))
#define MAX(a,b)		((a)>(b)?(a):(b))

#define PEAKS_RESOLUTION	5000

static char	verbose, printwindow;

static double_t freqResMag(const double_t *t,
	int len,
	int upoly,
	int phindex,
	double_t freq)
{
	double_t	resr = 0, resi = 0, coef, u;
	int		i;

	freq*=-2*M_PI;
	for (i=0; i<len; ++i)
	{
		/* res+= coef[idx]*cexp(idx*freq*I) */
		coef = t[i*upoly+phindex];
		u = i*freq;
		resr+= coef*cos(u);
		resi+= coef*sin(u);
	}
	/* return cabs(res);*/
	return hypot(resr,resi);
}

static double_t freqResFindSideHi(const double_t *t,
	int count,
	int upoly,
	int phindex,
	double_t freq,
	double_t prevMag)
{
	double_t		sideHi =-7000, freqMag = prevMag;

	while((freq+=1./PEAKS_RESOLUTION)<=.5)
	{
		freqMag = freqResMag(t,count,upoly,phindex,freq);
		if (freqMag>prevMag && freqMag>sideHi)
			sideHi = freqMag;
		prevMag = freqMag;
	}
	if (freq!=.5)
	{
		freqMag = freqResMag(t,count,upoly,phindex,.5);
		if (freqMag>prevMag && freqMag>sideHi)
			sideHi = freqMag;
	}
	return sideHi==-7000 ? freqMag:sideHi;
}

static void doFreqRes(const double_t *t,
	int count,
	int upoly,
	int phindex,
	double_t phase,
	double_t corner,
	double_t fresp,
	double_t fresa,
	bool ripple)
{
	double		sideHi =-7000, prevMag = 1e200, freqMag, freq;

	if (phase>=-1)
		printf("// phase %g:",
			phase);
	if (ripple)
	{
		double		rippleHi = sideHi, rippleLoPeak = prevMag;
		double		rippleLo = prevMag;
		double		ripLastPeakFreq = 0, ripLastZCrossFreq = 0;

		for(freq=0; freq<corner; freq+=.5/PEAKS_RESOLUTION)
		{
			freqMag = freqResMag(t,count,upoly,phindex,freq);
			if (freqMag>rippleHi)
				rippleHi = freqMag;
			if (freqMag<rippleLoPeak)
				rippleLoPeak = freqMag;
			if (freqMag>prevMag)
			{
				sideHi = freqMag, ripLastPeakFreq = freq;
				rippleLo = rippleLoPeak;
			}
			if (freqMag<1 && prevMag>1)
				ripLastZCrossFreq = freq;
			prevMag = freqMag;
		}
		rippleHi = log10(rippleHi)*20;
		rippleLo = log10(rippleLo)*20;
		printf(rippleLo>=4000 ? "\n// no ripple\n//":
"\n"
"// rippleHi %fdB, lo %fdB / dif %f\n"
"//",
			rippleHi+fresa,rippleLo+fresa,rippleHi-rippleLo);
		if (ripLastPeakFreq)
			printf(
"  last ripple peak freq %.4f, mag %fdB\n"
"//  last zero cross freq %.4f\n"
"//",
				ripLastPeakFreq,log10(sideHi)*20+fresa,ripLastZCrossFreq);
		sideHi =-7000;
	} else
		fputs("//",stdout);
	freqMag = freqResMag(t,count,upoly,phindex,corner);
	printf(" corner mag %gdB, ",
		log10(freqMag)*20+fresa);
	sideHi = freqResFindSideHi(t,count,upoly,phindex,corner,freqMag);
	printf(
"sideHi %.4fdB\n"
"// mag at %g: %gdB\n",
		log10(sideHi)*20+fresa,fresp,
		log10(freqResMag(t,count,upoly,phindex,fresp))*20+fresa);
}

/* first order modified bessel for Kaiser function */
double_t besselI0(double_t x);
double_t besselI0(double_t x)
{
#if 0 /* faster but less accurate version */
	double_t	y;

	x = fabs(x);
	if (x<3.75)
	{
		x*= 1/3.75, x*= x;
		return 1+x*(3.5156229+x*(3.0899424+x*(1.2067492+x*(0.2659732+x*(0.0360768+x*0.0045813)))));
	}
	y = 3.75/x;
	return (exp(x)/sqrt(x))*(0.39894228+y*(0.01328592+y*(0.00225319+y*(-0.00157565
		+y*(0.00916281+y*(-0.02057706+y*(0.02635537+y*(-0.01647633+y*0.00392377))))))));
#else
	double_t	y, a, p;
	int		i, f;

	x*= x, x*= .25, y = a = x;
	a*= x, a*= 1./4, y+= a;
	a*= x, a*= 1./9, y+= a;
	a*= x, a*= 1./16, y+= a;
	a*= x, a*= 1./25, y+= a;
	a*= x, a*= 1./36, y+= a;
	a*= x, a*= 1./49, y+= a;
	a*= x, a*= 1./64, y+= a;
	a*= x, a*= 1./81, y+= a, ++y;
	/* at this point covered for 80 bit when x'<=0.117786 */
	i = 10, f = 100;
	do
		p = y, a*= x/f, y+= a, f+= i, ++i, f+= i;
	while(p<y);
	return y;
#endif
}

static char *wintypes[] =
	{"rect","Bartlett","Hamming","Nuttall 3b","Nuttall 4c","Gaussian","Kaiser"};

/* returns 0 on error */
static int FIR_GenFilter(double_t *coef,
	double a, /* 2 * corner */
	double_t beta,
	double_t bbeta,
	double_t bbtri,
	double_t phase,
	double_t *rgain,
	int count,
	int type)
{
	double_t	lambda;
	double_t	smscale, smsize, sum, mid;
	int		i;

	if (count<=0)
		return 0;
	smsize = --count;
	if (type<0)// && phase<.5)
		type =~type, ++smsize;//, phase = 1-phase;
	mid = smsize/2.;
	if (verbose)
		printf("// window type %d (%s), smsize %g, mid %g\n",
			type,wintypes[type>5 ? 6:type],smsize,mid);
	if (printwindow)
		fputs("\ndouble window[] = {\n",stdout);
	if (type-2u>=5-2)
		smscale = 2./smsize;
	else
		smscale = 2*M_PI/smsize;
	lambda = M_PI*a;
	i = 0, sum = 0;
	do
	{
		double_t	x, w;
		/* phase must be applied to window for common frequency response */
		x = i+phase, w = smscale*x;
		if (x>smsize)
			w = 0;
		else
			switch(type)
			{
			default: /* Kaiser + optional triangular factor */
				--w, w = besselI0(beta*sqrt(1-w*w))/bbeta
					*(1-bbtri*fabs(w));
				break;
			case 5: /* Gaussian - 123-136db - 1-6 microdecibel pass ripple */
				w = fabs(w-1);
				if (w>=.9999999999999998)
					w = .9999999999999998;
				w = exp(-10.24*w*w); // *pow(1-w,1e-6);
				break;
			case 4: /* Nuttall 4c / 1.8687 bins W3 - 98-11x dB */
				w = .3635819-.4891775*cos(w)
					+.1365995*cos(2*w)
					-.0106411*cos(3*w);
				break;
			case 3: /* Nuttall 3b / 1.6162 bins W3 - +71 dB */
				w = .4243801-.4973406*cos(w)
					+.0782793*cos(2*w);
				break;
			case 2: /* Hamming - 38-53 dB */
				w = .54-.46*cos(w);
				break;
			case 1: /* Triangular / Bartlett - 26.5 dB */
				w = 1-fabs(w-1);
				break;
			case 0: /* rectangular / 1.0 bins zero - 13.259 dB PSLL */
				w = 1;
			}
		/* other window types, like Maxwell...kind of suck */
		if (printwindow)
			printf("% .15f,\n",
				w);
		x-= mid;
		sum+= coef[i] = (x*=lambda)==0 ? w:sin(x)/x*w;
	} while(++i<=count);
	*rgain = sum;
	if (printwindow)
	{
		fputs("};\n\n",stdout);
		printwindow = 0;
	}
	return 1;
}

/* returns <=0 on error */
static int FIR_PrintFilters(double polyo,
	int numcoef,
	double_t beta,
	double_t btri,
	double_t corner,
	double_t scale,
	int outtype,
	bool group,
	bool ncx,
	double_t fresp,
	bool fresr)
{
	double_t	*coefbuf, a, mmax;
	double_t	mag, bbeta, rnd;
	int		c, r, fnum, type, nfilt, qsum, iqs;
	int		dir = polyo>=0, dnc, hnc, dnf, lek, groupi;
	int		dgc = outtype ? 15:16;

	if (!dir)
		polyo =-polyo;
	iqs = 0;
	if (outtype>=2e7)
		scale = exp2(outtype*1e-7), iqs = (int)(scale+.5);
	lek = iqs-1u<1000 ? iqs<100 ? 15:7:3;
	a = corner*2;
	if ((coefbuf=malloc(sizeof(*coefbuf)*numcoef))==NULL)
		return -2;
	nfilt = 1;
	if (polyo>=1)
		nfilt = polyo, polyo = 0;
	type = beta*10+.5;
	if (ncx)
		type =~type;
	bbeta = besselI0(beta);
	fnum = dir ? 0:nfilt;
	dir = dir+dir-1;
	/* end-coef pretty useless in q mode */
	dnc = numcoef+((dir|-iqs)>>31);
	hnc = dnc&1 ? -9:(dnc-1)>>1;
//printf("%d\n",hnc);
	dnf = nfilt+(dir>>31);
	c = groupi = 0;
	if (outtype-1u>1)
		printf(outtype>2 ? nfilt>1 ? "\ncoef[%d * %d] =\n{":"\ncoef[%d] =\n{"
			:btri ? "// %d taps * %d / bt=%.9g*(1-%.7g|n|)\n"
			:"// %d taps * %d / bt=%.9g\n",
			dnc,nfilt,beta,btri);
	goto start;
	do
	{
		double_t	v, phase;
		int		i, j, u, mxu, ovun;

		if (!iqs)
			fputc('\n',stdout);
	start:	mmax =-1e9; //mmin = 1e9;
		phase = polyo+(double_t)fnum/nfilt;
		if ((r=FIR_GenFilter(coefbuf,a,beta,bbeta,btri,
				phase,&mag,dnc/*numcoef*/,type))==0)
			break;
		mag = 1/mag;
		rnd = .5, mxu = ovun = 0;
		if (iqs)
		{ /* in quantize mode, correct missed sum by adjusting round */
			double		drop = iqs>128 ? .0333:.0218;

			j = 74;
			do
			{
				i = 0, qsum = 0, mxu = 0;
				do
				{
					v = coefbuf[i]*mag*scale;
					v+= v<0 ? -rnd:rnd;
					u = (int)v;
					if (u>mxu)
						mxu = u;
					qsum+= u;
				} while(++i<dnc);
				ovun = qsum-iqs;
				if (!ovun)
					break;
				rnd+= ovun<0 ? drop:-drop;
				drop*= .901;
				if (j==20)
					mag+= mag/qsum*.25;
			} while(--j);
		}
		if (fresp>=0 || fresr)
		{
			doFreqRes(coefbuf,dnc,1,0,phase,corner,fresp,log10(mag)*20,fresr);
			fputc('\n',stdout);
		}
		i = 0, qsum = 0, mag*= scale;
		do
		{
			v = coefbuf[i]*mag;
			if (iqs)
			{
				//if (v<mmin && (int)v)
				//	mmin = v;
				if (v>mmax)
					mmax = v;
				v+= v<0 ? -rnd:rnd;
				u = (int)v;
				if (u && fnum==nfilt)
					if (ovun==-1 && i==hnc)
						++u;
					else if ((ovun==1 && (i-hnc+1u)<3u))
						u+= u==mxu ? 1:-1;
				qsum+= u;
				v = u;
			}
			if (fabs(v)<6e-17)
				v = copysign(0,v);
			if (outtype==1) /* native data */
				fwrite(&v,8,1,stdout);
			else if (outtype==2)
			{ /* swapped data */
				fputc(((unsigned char*)&v)[7],stdout);
				fputc(((unsigned char*)&v)[6],stdout);
				fputc(((unsigned char*)&v)[5],stdout);
				fputc(((unsigned char*)&v)[4],stdout);
				fputc(((unsigned char*)&v)[3],stdout);
				fputc(((unsigned char*)&v)[2],stdout);
				fputc(((unsigned char*)&v)[1],stdout);
				fputc(((unsigned char*)&v)[0],stdout);
			} else
				printf(outtype>2 ? i&lek ? "%.*g,"
					:groupi ? " %.*g,":"\n\t%.*g,":"% .*f\n",
					dgc,v);
		} while(++i<dnc);
		if (iqs)
		{
			if (qsum!=iqs)
				printf(" /*qsum=%d,max=%.1f*/",
					qsum,round(mmax*10)*.1);
			qsum = 0;
		}
		groupi^= group;
		fnum+= dir;
	} while(++c<nfilt);
	free(coefbuf);

	if (outtype>2)
		fputs("\n};\n",stdout);
	return r>0 ? numcoef:r;
}

static int typeToTaps(double_t beta,
	double_t tband,
	double_t Fs,
	int odd)
{
	double_t	dbins;
	int		t = beta*10+.5;

	dbins = t>5 ? (beta<1.5 ? 1.5:beta)*(2/M_PI)
		:t<1 ? .91
		:t<3 ? 3.4
		:t<4 ? 5.6
		:t<5 ? 7.8
		:14.2;
	t = (int)ceil(dbins*Fs/tband)+1;
	return odd==2 ? t&~1:t|odd;
}


/*
 Dolph Chebyshev windowing

for all-pass corner 0.375, maximum filter atten.
3	5.3789191269869
5	11.35233308819
7	17.65662875398
9	24.1434423003
11	30.7584039832
13	37.469938989
15	44.253659601
17	51.090217885
19	57.96495447
21	64.86723733
23	71.78956218
25	78.72669164
27	85.6749658
29	92.63180093
31	99.59534396
33	106.56424059
35	113.5374811
37	120.514298
39	127.49410002
41	134.4764102
43	141.46087

quadratic fit (x<8): x(.0413602130737x+2.655825276011)-2.960798618709
quadratic fit (8<x<16): x(2.883649333+x(.02883808-.0005080734x))-3.7749006683

linear fit (x<4): .4567337148736483x+.5432662851263517
cubic fit (a<37): x(x(x*1.246838661e-5-.0013605701199)+.348719548165)+1.1983219301
sextic to taps (a<134.47641):
 x(x(x(x(x(x*-6.993196e-13+4.11216e-10)-1.0251089e-7)+1.4047054e-5)-.0011421219)+.34048227)+1.28032742
linear fit (a>134.47641)
 .28634999098*x+2.4920656093
*/

static double_t cheby_poly(double_t n,
	double_t x)
{
	return fabs(x)<=1 ? cos(n*acos(x)):cosh(n*acosh(x));
}

/* Trial & error has proven other methods - including front/back partials, FIR
  phase adjustment (above) - don't work better than the implementation below.
  The hacky N+=1 globally to fit the appearance of 0.5 is also incorrect.
*/
static void cheby_window(double_t *w,
	int N,
	bool ncx,
	double_t M,
	double_t Mp,
	double_t tg,
	double_t btri)
{
	double_t	x0, n, Nf, countf, sum, max;
	int		i, count, nn, dir;

	M+= ncx;
	max = sum = 0;
	Mp = M-Mp;
	dir = Mp>.5;
	if (dir) /* best results are in 0-0.5, so mirror and flip in 0.5-1 */
		M-= Mp, Mp = 1-Mp, M+= Mp;
	Mp+= Mp;
	if (ncx)
	{
		Nf = N-Mp;
		countf = Nf-1;
		M-= Mp, Mp = 0;
		count = N-2;
	} else
	{
		Nf = N+Mp;
		countf = Nf-1;
		count = (int)countf;
	}
	nn = count;
	dir^= ncx;
	x0 = cosh(acosh(tg)/(Nf-1));
	if (verbose==1)
		printf(
"// Dolph-Chebyshev, tg=%.10g, x0=%.10g\n"
"//  N=%.16g, c=%.16g, M=%g/%g\n",
			tg,x0,Nf,countf,M,Mp);
	do
	{
		n = 2*(nn-M);
		for(i=1,sum=0; i<=M; ++i)
			sum+= cheby_poly(Nf-1,x0*cos(M_PI*i/Nf))*cos(n*M_PI*i/Nf);
		sum+= sum, sum+= tg;
		if (!max)
			max = sum;
		sum*= 1-btri*fabs(nn-countf*.5)/countf; //*n;
		sum/= max;
		w[dir ? count-nn:nn] = sum;
	} while(--nn>=0);
}

static void print_chebyfilter(int N,
	double_t atten,
	double_t fatten,
	double_t corner,
	double_t plyo,
	double_t btri,
	double_t scale,
	bool cdv2,
	int outtype,
	bool group,
	bool ncx,
	unsigned char autoattempts,
	double_t fresp,
	bool fresr)
{
	double_t	*w, *t;
	double_t	lambda, phase, phasea, sum, max, tg, mag, rnd;
	double_t	aatten = fabs(atten), nhatten;
	int		iqs, qsum, u;
	int		i, j, k, poly, upoly, count;//, hasfrac;
	int		groupi;
	bool		pscaled = plyo<=-1, adjust, attendir = autoattempts;
	bool		adjusted = 0;

	if (corner>0.5)
		corner = 0.5;
	N+= ncx;
	phase = sum = max = qsum = 0;
	iqs = 0;
	if (outtype>=2e7)
		scale = exp2(outtype*1e-7), iqs = (int)(scale+.5);
	if (pscaled)
		plyo =-plyo;
	groupi = 0;
	poly = 1;
	if (plyo<1)
		phase = plyo;//, hasfrac = !!plyo && !ncx;
	else if (plyo>1)
	{
		//++hasfrac;
		if (plyo<256)
			poly = plyo;
		else
			poly = 256;
	}
	upoly = poly;
	if (pscaled)
		N*= poly, corner/= poly, upoly = 1;
	phasea = (N-1)*.5;
	phase+= phasea-ncx;
	lambda = corner*M_PI*2;
	if (fatten>0)
		atten =-fatten;
	tg = pow(10,fabs(atten)/20);  /* 1/r term [2], 10^gamma [2] */
	w = calloc(N+1,sizeof(double_t));
	t = calloc((N+1)*upoly,sizeof(double_t));

	cheby_window(w,N,ncx,phase,phasea,tg,btri);
	verbose+= verbose;

	count = N-ncx;
	printf("\n// corner %g, %d taps\n",
		corner,count);
	nhatten = atten;
	adjust = atten<0;
	if (autoattempts<80)
		autoattempts = 80;
	//for(j=upoly; --j>=0; )
	for(j=0; j<upoly; ++j)
	{
		double_t	x = phase+(double_t)j/upoly, pmag;
		double_t	dif, ldif, pdif, latten, lmag, difadj;
		int		attempt, arev;
		bool		edge;

		mag = 0, lmag = pmag = ldif = dif = pdif = 7000;
		if (adjust)
		{
			printf("\n// adjusting %f for sidelobes..",
				x-phasea+ncx,fatten);
			atten = fatten>0 ? -fatten
				:(!(upoly&1) && j==(upoly>>1)) ? nhatten
				:aatten*-.99;
			adjusted = 1;
			if (j)
				tg = pow(10,fabs(atten)/20);
		}
		latten = atten;
		//for(trial=0; trial<2 && dif>.1; ++trial)
		for(attempt=arev=0,edge=0,difadj=1; attempt<=autoattempts+2;
				++attempt)
		 {
			cheby_window(w,N,ncx,x,phasea,tg,btri);
			for(i=0,sum=0; i<count; ++i)
			{
				double_t	n = lambda*(i-x);

				n = n ? sin(n)/n:1;
				sum+= t[i*upoly+j] = n*w[i];
			}
			for(i=0; i<count; ++i)
				t[i*upoly+j]/= sum;
			if (atten>0 || attempt>autoattempts+1)
				break;
			mag = log10(freqResFindSideHi(t,count,upoly,j,corner,
				freqResMag(t,count,upoly,j,corner)))*20;
			pdif = dif;
			dif = mag+aatten;
			if (attendir && mag<-aatten)
				dif*= fatten>0 ? .0001:.25;
			if (fabs(dif)<fabs(ldif) || (attendir && mag<-137
					&& (!(fatten>0) || mag<-aatten)))
			{
				ldif = dif, latten = atten, lmag = mag;
				//printf("\nfound lower\n");
				if (fatten>0 && attendir && -mag>aatten)
					aatten =-mag-1;
			} else if (attempt>autoattempts)
			{ /* reaching end of trial and error; go with best */
				atten = latten;
				goto punt;
			}
			if (verbose)
				printf("\n%g/%g/%.13g",
					mag,dif,atten);
			if (attendir && mag<-299)
			{
				mag = (mag+300)*.01;
				if (attempt>5 && attempt<(autoattempts*3>>2))
					attempt = autoattempts*3>>2;
			} else if (fabs(mag+=aatten)<2e-12)
				break;
			if (fatten>0 && (attempt>10 || aatten<120))
				mag*= .01;
			if (!(attempt&3) && !verbose)
			{
				fputc('.',stdout);
				fflush(stdout);
			}
			atten-= (arev ? attempt>62 ? mag*(arev&2 ? -.001:.001)
				:attempt>54 ? mag*(arev&2 ? -.01:.01)
				:attempt>45 ? mag*(arev&2 ? -.03:.03)
				:attempt>36 ? mag*(arev&2 ? -.06:.06)
				:attempt>23 || arev>1 ? mag*(arev&2 ? -.12:.12)
				:mag*(arev&2 ? -.25:.25)
				:mag*.75)*difadj;
			if (attempt>3 && fabs(pmag)<fabs(mag))
			{
				++arev; /* only incrementing one vs two to give some time */
				dif = ldif, atten = latten;
				if (attempt>8)
					difadj*= .95;
			}
			if (atten<-192 || fabs(atten)<aatten*.2 || atten>=0)
			{
				if (!arev)
				{
					if (!edge)
						printf(
"\n//  hit ceiling; adjusting for edge..",
							stdout);
					arev = 3, edge = 1;
					atten = fatten>0 ? -fatten:-aatten;
				} else
				{
					if (aatten>120 && autoattempts<81)
					{ /* if it was way high, start over */
						aatten = 120, fatten = 0;
						attempt = arev = 0;
						if (verbose)
							fputs(
"\n//  hit limit; restarting with 120",
							stdout);
						goto punt;
					}
					fputs(
"\n// ** could not find good adjust **",
						stdout);
					adjusted = 0;
					atten = aatten;
					attempt = autoattempts+1;
					if (!(upoly&1) && j==(upoly>>1))
					{ /* copy an adjusted value from last */
					//printf("+-+- %g\n",
					// nhatten);
						atten = nhatten;//*.9921875;
						lmag = 0;
					} else if (fatten>0)
						atten =-fatten;
				}
			} else if (verbose)
				printf(", trying %g\n",
					atten);
		punt:	tg = pow(10,fabs(atten)/20);
			pmag = mag;
		 }

		if (atten<0)
		{
			/* one last test */
			if (attendir && lmag<-aatten)
			{
				atten = latten;
				tg = pow(10,atten/-20);
				cheby_window(w,N,ncx,x,phasea,tg,btri);
				for(i=0,sum=0; i<count; ++i)
				{
					double_t	n = lambda*(i-x);

					n = n ? sin(n)/n:1;
					sum+= t[i*upoly+j] = n*w[i];
				}
				for(i=0; i<count; ++i)
					t[i*upoly+j]/= sum;
			}

			printf(adjusted ? "\n//  adjusted to %.10g\n":"\n//  set to %.10g\n",
				-atten);
			if (j==(upoly>>1)-1)
				nhatten = atten;
		}
		if (printwindow)
		{
			fputs("\ndouble window[] = {\n",stdout);
			for(i=0; i<count; ++i)
				printf(/*j&3 ? "%.15f,":*/ "% .15f,\n",
					w[i]);
			fputs("};\n\n",stdout);
			printwindow = 0;
		}
		if (fresp>=0 || fresr)
			doFreqRes(t,count,upoly,j,adjust ? -2:x-phasea+ncx,
				corner,fresp,0,fresr);
	}
	mag = 1, rnd = .5;
	if (poly==1 && iqs)
	{ /* in quantize mode, correct missed sum by adjusting round */
		double		v, drop = iqs>128 ? .0333:.0218;
		int		ovun;

		j = 74;
		do
		{
			i = 0, sum = 0;
			do
			{
				v = t[i]*mag*scale;
				v+= v<0 ? -rnd:rnd;
				u = (int)v;
				sum+= u;
			} while(++i<count);
			ovun = sum-scale;
			if (!ovun)
				break;
			rnd+= ovun<0 ? drop:-drop;
			drop*= .901;
			if (j==20)
				mag+= mag/scale*.25;
		} while(--j);
	}

	fputs("double filter[] =\n{\n",stdout);
	count*= upoly, mag*= scale;
	for(k=0; k<poly; ++k,--count)
	{
		for(i=k,j=0,sum=0; i<count; ++i)
		{
			sum+= t[i];
			if (pscaled && ++j<poly)
				continue;
			if (cdv2)
			{
				++i;
				if (i<count)
					sum+= t[i];
			}
			sum*= mag;
			if (iqs) //(mag)
			{
				sum+= sum<0 ? -rnd:rnd;
				u = (int)sum;
				qsum+= u;
				printf(j || i%upoly ? " % d,":"% d,",
					u);
			} else
				printf(j || i%upoly ? " % .15f,":"% .15f,",
					sum);
			putc((groupi^=group) ? ' ':'\n',stdout);
			j = 0, sum = 0;
		}
		if (!pscaled || poly<2)
			break;
		if (iqs)
		{
			sum*= scale, sum+= sum<0 ? -.5:.5;
			u = (int)sum;
			qsum+= u;
			printf("% d, //%d\n",
				u,k);
		} else
			printf("% .15f, //%d\n",
				sum*scale,k);
	}
	fputs("};\n\n",stdout);
	if (poly==1 && iqs && qsum!=iqs)
		printf("// qsum=%d\n",
			qsum);
	free(w);
	free(t);
}

static int attenToTaps(double_t a,
	double_t Ft,
	double_t Fc,
	int odd)
{
	//int		n = 0;

	if (a<0)
		a =-a;//, n =-1;
	/* apply polynomial for count reflecting all-pass */
	if (a<37)
		a = a*(a*(a*1.246838661e-5-.0013605701199)+.348719548165)+1.1983219301;
	else if (a<134.47641)
		a = a*(a*(a*(a*(a*(a*-6.993196e-13+4.11216e-10)-1.0251089e-7)+1.4047054e-5)
			-.0011421219)+.34048227)+1.28032742;
	else
		a = .28634999098*a+2.4920656093;
	if (Fc<.375)
		a+= .375-Fc;
	/* scale and round to transition band */
	return ((int)(a*(8*3./100)/Ft+.999999)&~1)+odd;//)^n)+n;
}

int main(int argc,
	char **argv)
{
	int		r, ncoef = 0, otyp = 0, odd = 1;
	char		*s, *ffs, fresr = 0;
	double_t	corner, Fs = 1, tband = 0, bt = 9.3, btri = 0;
	double_t	plyo = 0, fresp =-1, scale = 1;
	bool		group = 0, rep, tb = 0;

	if (argc<=0 || !argv || !*argv)
		return 0;
	do
	{
		rep = 0;
		if (argc>1 && (s=argv[1])!=NULL)
			if (!strcmp(s,"MSB"))
				{otyp = 1+*(char*)&odd; goto reo;}
			else if (!strcmp(s,"LSB"))
			{
				otyp = 2-*(char*)&odd;
			reo:
			#ifdef _O_BINARY
				_setmode(_fileno(stdout),_O_BINARY);
			#else
				freopen("stdout.bin","wb",stdout);
			#endif
				goto acp;
			} else if ((*s=='C' || *s=='c') && !s[1])
				{otyp = 3; goto acp;}
			else if (!strcmp(s,"fract") || !strcmp(s,"Fract"))
			{
				otyp = 30e7;
				goto acp;
			} else if (*s=='s')
			{
				++s;
				scale = atof(s);
				goto acp;
			} else if (*s=='q')
			{
				if (*++s=='g')
					group = 1, ++s;
				tband = atof(s); /* using temp variable */
				if (tband<2 || tband>30)
				{
					fprintf(stderr,"### bad quantize parameter '%s'\n",
						s);
					return 1;
				}
				otyp = tband*1e7;
				goto acp;
			} else if (!memcmp(s,"-fres",5) && ((fresr=s[5])==0 || fresr=='r'))
			{
				if (--argc)
				{
					++argv;
					fresp = strtod(argv[1],&s);
					if (*s=='/')
						fresp/= atof(s+1);
				}
				rep = 1;
				goto acp;
			} else if (!strcmp(s,"-w"))
			{
				printwindow = 1;
				rep = 1;
				goto acp;
			} else if (!strcmp(s,"-v") || !strcmp(s,"-verbose"))
			{
				verbose = 1;
				rep = 1;
			acp:	++argv, --argc;
			} else
			{
				if (*s=='-')
					++s;
				if (!strchr(".0123456789",*s) && strcmp(s,"text"))
				{
					fprintf(stderr,"### unknown parameter '%s'\n",
						s);
					return 1;
				}
			}
	} while(rep);
	if (argc<3)
	{
		fprintf(stderr,
"# usage: mkfilter [outtype] corner[/Fs] ncoef [wtype|beta] [nfilt/phase]\n"
"#\n"
"# ncoef         append tb for transition band approximation\n"
"#                append e to prefer even coefs\n"
"#                append x to prefer extended window sizing\n"
"# beta|wtype    0=rectangular; 0.1=triangular window; 0.2=Hamming\n"
"#                0.3=Nuttall-3; 0.4=Nuttall-4; 0.5=Gaussian (123+dB)\n"
"#                >0.55 Kaiser/beta (default:9.3)\n"
"#                 for alpha, append pi; for Kaiser's dB formula, append dB\n"
"# outtype       LSB=little endian data out; MSB=big endian data\n"
"#                 C=C formatted; fract=quantize integer 30 bits\n"
"#                 q#=quantize int in # bits (2.0-30 only); qg# to group 2x\n"
"#                 s#=output scale (floating point)\n"
"#                 default:text\n"
"# -fres n[/Fs]  find frequency response info for freq n and peaks\n"
"#                 fresr=also find ripple peaks\n"
"# -v[erbose]    verbose printout\n"
"# -w            printf window\n"
"#\n"
"# for Dolph Chebyshev window, parameters are instead:\n"
"#    ntaps[>=2|tb] [[-]attendB[+]] [corner] [polyc/phase]\n"
"\n"
			);
		return 0;
	}
	s = argv[1]+strlen(argv[1]);
	if ((corner=strtod(argv[1],&s))<=0 || argv[1]==s || !s)
		goto badcor;
	tband = corner;
	if (*s=='t' && s[1]=='b')
		s+= 2, tb = 1;

	if (corner>=2 || tb)
	{
		double_t	atten = 100, fatten = 0;
		double_t	btri = 0;
		bool		cdv2 = 0;
		unsigned char	autoattempts = 0;

		if (*s=='x' && (!s[1] || s[1]=='/'))
			++s, odd = 2;
		else if (!tb && *s=='/' && s[1]=='2' && !s[2])
			cdv2 = 1;
		else if (*s && *s!='/')
		{
			fputs("### bad ncoef/tran",stderr);
			return 1;
		}
		if (*s=='/')
		{
			++s;
			ffs = s+strlen(s);
			if (!((Fs=strtod(s,&ffs))>0 && Fs<INFINITY) || s==ffs)
				goto burp;
		}
		if (argc>2)
		{
			s = argv[2]+strlen(argv[2]);
			if ((atten=strtod(argv[2],&s))==0 || !(atten<=308)
					|| argv[2]==s || !s)
				goto invat;
			if (*s=='+')
			{
				++s, autoattempts = 1;
				if (*s=='+')
					++s, autoattempts = 240;
			}
			if (*s==':')
			{
				fatten = strtod(s+1,&s);
				if (!s)
					goto invat;
			}
			if (*s=='*')
			{
				if (!(fabs(btri=atof(++s))<1))
				{
					fprintf(stderr,"### bartlet adjustment must be in (-1,1)\n");
					return 1;
				}
			} else if (*s && *s!=' ')
			{
			invat:	fputs("### invalid atten",stderr);
				return 1;
			}
		}
		tband/= Fs;
		if (tb)
			ncoef = attenToTaps(fatten>0 ? fatten:atten,tband,corner,odd!=2);
		else
			ncoef = (int)tband;
		if (!(ncoef>0 && ncoef<65536))
		{
			fputs("### ncoef out of range",stderr);
			return 1;
		}
		if (argc>3)
		{
			tband = 0;
			if (!((corner=strtod(argv[3],&s))<=.5 && corner>=0)
					|| argv[3]==s || !s
					|| (*s && *s!=' '))
			{
				fputs("### invalid corner\n",stderr);
				return 1;
			}
			corner/= Fs;
		} else
			corner = 0;
		if (corner<=0)
		{
			fputs("// defaulting to max corner",stdout);
			corner = (1-tband)*.5;
		}
		if (argc>4 && !isfinite(plyo=atof(argv[4])))
		{
			fputs("### invalid phase offset/poly count\n",stderr);
			return 1;
		}
		print_chebyfilter(ncoef,atten,fatten>0 ? fatten:0,corner,plyo,btri,
			scale,cdv2,otyp,group,odd==2,autoattempts,fresp,fresr);
		return 0;
	}
	/* generic window types */
	if (s)
	{
		while(*s==' ' || *s=='\t')
			++s;
		if (*s && *(unsigned char*)s>' ')
		{
			if (*s++!='/')
			{
				fprintf(stderr,"### unknown corner parameter '%s'\n",
					s-1);
				return 1;
			}
			ffs = s+strlen(s);
			if (!((Fs=strtod(s,&ffs))>0 && Fs<INFINITY) || s==ffs)
			{
			burp:	fprintf(stderr,"### bad rate parameter\n");
				return 1;
			}
			corner/= Fs;
		}
	}
	if (corner>.5)
	{
	badcor:	fprintf(stderr,Fs!=1 ? "### bad corner or rate param (resulted in %.15g)\n"
				:"### bad corner parameter (not in (0,.5])\n",
			corner);
		return 1;
	}
	s = argv[2]+strlen(argv[2]);
	if ((tband=strtod(argv[2],&s))<=0 || argv[2]==s || !s)
		goto badtb;
	if (*s!='t' || *++s!='b' || (*++s && *s!='x'))
	{
		if (*s=='x' && !s[1])
			{odd = 2; goto dir;}
		else if (*s=='e' && !s[1])
			odd = 0;
		else
		{
			if (*s)
			{
			badtb:	fprintf(stderr,"### bad ncoef/trans parameter\n");
				return 1;
			}
			/* explicit coeficient count */
		dir:	if ((ncoef=tband)<3)
			{
				fprintf(stderr,"### bad ncoef parameter\n");
				return 1;
			}
		}
	} else if (*s=='x' && !s[1])
		odd = 2;
	if (argc>3)
	{
		s = argv[3]+strlen(argv[3]);
		if ((bt=strtod(argv[3],&s))<0 || argv[3]==s || !s)
			goto badbt;
		if (*s=='p' && *++s=='i' && (!s[1] || s[1]=='*'))
		{ /* bt is requested to be an alpha parameter instead */
			bt*= M_PI, ++s;
			if (bt<.5501)
			{
				fprintf(stderr,"### unsupported alpha\n");
				return 1;
			}
		} else if (*s=='d' && (s[1]=='b' || s[1]=='B') && (!s[2] || s[2]=='*'))
		{ /* Kaiser's decibel approximation formula */
			s+= 2;
			if (bt>50)
				bt = .1102*(bt-8.7);
			else if ((bt-=21)<0 || (bt=.5842*pow(bt,.4)+.07886*bt)<.5501)
				bt = .5501;
		} else if (*s && *s!='*')
		{
		badbt:	fprintf(stderr,"### bad type|beta parameter\n");
			return 1;
		}
		if (*s=='*')
			if (!(fabs(btri=atof(++s))<1))
			{
				fprintf(stderr,"### bartlet adjustment must be in (-1,1)\n");
				return 1;
			}
		if (bt>=50*M_PI)
		{
			fprintf(stderr,"### beta is too high\n");
			return 1;
		}
	}
	if (argc>4 && fabs(plyo=atof(argv[4]))>256) /* more than 256 exceeds 16 bit error */
	{
		fprintf(stderr,"### bad nfilt/offs parameter\n");
		return 1;
	}
	if (!ncoef)
	{ /* transition band */
		if (tband>Fs*.5)
		{
			fprintf(stderr,"### transition band must be in (0,Fs/2]\n");
			return 1;
		}
		ncoef = typeToTaps(bt,tband,Fs,odd);
		if (ncoef>1000000)
		{
			fprintf(stderr,"### coeficient count>1000000\n");
			return 1;
		}
	}
	r = FIR_PrintFilters(plyo,ncoef,bt,btri,corner,scale,otyp,group,odd==2,
		fresp,fresr);
	if (r<=0)
	{
		fprintf(stderr,r==-2 ? "### out of memory\n":"### error #%d\n",
			r);
		return 1;
	}
	return 0;
}
